﻿using System;
using System.Linq;
using System.Collections.Generic;

using Rays;

using Raytracer;
using Raytracer.Math;
using Raytracer.Primitives;

namespace Rays {

    public class ConstantSurface : Surface {
        private readonly double m_Roughness, m_Reflection;
        private readonly Color m_Diffuse, m_Specular;

        public ConstantSurface(Color diffuse,
                               Color specular,
                               double reflection,
                               double roughness) {
           this.m_Diffuse = diffuse;
           this.m_Specular = specular;
           this.m_Reflection = reflection;
           this.m_Roughness = roughness;
        }

        public Color Diffuse(Vector3 value) {
            return m_Diffuse;
        }

        public double Reflection(Vector3 value) {
            return m_Reflection;
        }

        public double Roughness(Vector3 value) {
            return m_Roughness;
        }

        public Color Specular(Vector3 value) {
            return m_Specular;
        }
    }

    public class Checkerboard : Surface {
        public static readonly Checkerboard Instance = new Checkerboard();
        private Checkerboard() { }

        public Color Diffuse(Vector3 pos) {
            return ((int)(Math.Floor(pos.z) + Math.Floor(pos.x)) % 2 != 0)
                   ? new Color(1, 1, 1)
                   : new Color(0, 0, 0);
        }

        public double Reflection(Vector3 pos) {
            return ((int)(Math.Floor(pos.z) + Math.Floor(pos.x)) % 2 != 0)
                   ? 0.1
                   : 0.7;
        }

        public double Roughness(Vector3 value) {
            return 150;
        }

        public Color Specular(Vector3 value) {
            return new Color(1, 1, 1);
        }
    }

    public class SphereGenerator : IPrimitiveFactory {
        public string Type {
            get { return "Sphere"; }
        }

        public Primitive Create(Ast data, Context context) {
            var props = data.Eval<List>(context);
            return new Sphere(props.GetField("m", context).ToVector(context),
                              props.GetField("r", context).Eval<Number>(context).Value);
        }
    }

    public class PlaneGenerator : IPrimitiveFactory {
        public string Type {
            get { return "Plane"; }
        }

        public Primitive Create(Ast data, Context context) {
            var props = data.Eval<List>(context);

            var d = props.GetField("d", context).Eval<Number>(context);
            var n = props.GetField("n", context).ToVector(context);

            return new Plane(n,
                             d.Value);
        }
    }
 
}