﻿namespace Raytracer.Math

type SysColor = System.Drawing.Color

type Vector3(x : float, y : float, z : float) = 
    member this.x = x
    member this.y = y
    member this.z = z
    
    member this.LengthSq = x*x + y*y + z*z
    member this.Length = this.LengthSq ** 0.5
    member this.Unit = this * (1.0 / this.Length)
    
    static member (+) (a : Vector3, b : Vector3) = new Vector3(a.x + b.x, a.y + b.y, a.z + b.z)
    static member (-) (a : Vector3, b : Vector3) = new Vector3(a.x - b.x, a.y - b.y, a.z - b.z)
    static member (~-) (a : Vector3) = -1.0 * a
    
    static member (*) (r : float, a : Vector3) = new Vector3(r * a.x, r * a.y, r * a.z)
    static member (*) (a : Vector3, r : float) = r * a  
    
    static member (.*) (a : Vector3, b : Vector3) = a.x * b.x + a.y * b.y + a.z * b.z
    static member (.**) (a : Vector3, b : Vector3) = new Vector3((a.y * b.z) - (a.z * b.y), (a.z * b.x) - (a.x * b.z), (a.x * b.y) - (a.y * b.x))
    
    override this.ToString() = (x, y, z).ToString()

type Color(r : float, g : float, b : float) = 
    let threshold n = 255.0 * (if n > 1.0 then 1.0 else n) |> int
    
    member this.R = r
    member this.G = g
    member this.B = b    
    
    static member (+)  (a : Color, b : Color) = new Color(a.R + b.R, a.G + b.G, a.B + b.B)
    static member (-)  (a : Color, b : Color) = new Color(a.R - b.R, a.G - b.G, a.B - b.B)
    static member (.*) (a : Color, b : Color) = new Color(a.R * b.R, a.G * b.G, a.B * b.B)
    
    static member (*) (r : float, a : Color) = new Color(r * a.R, r * a.G, r * a.B)
    static member (*) (a : Color, r : float) = r * a  
    static member Zero = new Color(0.0, 0.0, 0.0)
    
    member this.ToColor = System.Drawing.Color.FromArgb(255, threshold r, threshold g, threshold b)
    
type Ray(p : Vector3, u : Vector3) = 
    let u0 = u.Unit
    
    member this.Base = p
    member this.Direction = u0
 
    member this.At t = this.Base + t * this.Direction
    static member FromPoints (A : Vector3, B : Vector3) = new Ray(A, B - A)